/*
 *  This file if part of nzbget
 *
 *  Copyright (C) 2004  Sven Henkel <sidddy@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Revision: 1.9 $
 * $Date: 2004/07/13 21:52:50 $
 *
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <unistd.h>

#include "ArticleDownloader.h"
#include "Decoder.h"
#include "global.h"
#include "Frontend.h"

#include "MultiServerPool.h"

#include "NZBQueueCoordinator.h"

extern NZBQueueCoordinator* g_pNZBQueueCoordinator;
extern Frontend* frontend;
extern Options* options;
extern float g_fDownloadRate;
extern bool g_bPause;

ArticleDownloader::ArticleDownloader(ArticleJob* pJob, char* szTempDir)
{
	m_pJob = pJob;
	m_pJob->setStatus(JOBSTATUS_ASSIGNED);
	m_bCancelDownload = false;

	if( szTempDir )
	{
		m_szTempDir = (char*) malloc(strlen(szTempDir) + 2 );
		memset( m_szTempDir, 0, strlen(szTempDir) + 2 );
		memcpy( m_szTempDir, szTempDir, strlen(szTempDir) );
		if( m_szTempDir[strlen(m_szTempDir)-1] != '/')
		{
			m_szTempDir[strlen(m_szTempDir)] = '/';
		}
	}
	else
	{
		m_szTempDir = strdup("./");
	}
}

ArticleDownloader::~ArticleDownloader()
{
	if( m_szTempDir )
	{
		free( m_szTempDir );
	}
}

void ArticleDownloader::run()
{
	// Pre-init
	bool bThrottleDownloadRate = false;
	if( g_fDownloadRate > 0.0f )
	{
		bThrottleDownloadRate = true;
	}

	m_pJob->setStatus( JOBSTATUS_RUNNING );
	
	int level = 0;
	while( !m_bCancelDownload )
	{
		NNTPConnection* pConnection = MultiServerPool::getInstance()->getConnection( level );

		if(!pConnection)
		{
			// all pConnectionection-levels passed, still not finished...
			// seems like we failed.
			m_pJob->setStatus(JOBSTATUS_FAILED);
			error("%s FAILED! Article unavailable on all servers.", m_pJob->getMessageID());
			break;
        	}

		// Okay, we got a pConnectionection. Now start downloading!!

		// but at first, change group! dryan's server wants it this way... ;-)
		bool grpchanged = true;

		for (unsigned int i = 0; i < m_pJob->getGroups()->size(); i++)
		{
			if (pConnection->joinGroup((*(m_pJob->getGroups()))[i]))
			{
				grpchanged = false;
			}
			else
			{
				grpchanged = true;
				break;
			}
		}

		if (grpchanged)
		{
			// now, let's begin!
			char tmp[1024];
      
			snprintf(tmp, 1024, "ARTICLE %s\r\n", m_pJob->getMessageID());

			char* answer = NULL;

			for (int retry = 3; retry > 0; retry--)
			{
				answer = pConnection->request(tmp);
				if (answer && (!strncmp(answer, "2", 1)))
				{
					break;
				}
			}

			if (answer && (!strncmp(answer, "2", 1)))
			{
				// positive answer!
				g_pNZBQueueCoordinator->JobStarted( m_pJob, pConnection );
              
				char* line = NULL;

				char tmpfilename[1024];

				snprintf(tmpfilename, 1024, "%snzbgettemp-XXXXXX", m_szTempDir);

				int tmpfd = mkstemp(tmpfilename);

				FILE* outfile = fdopen(tmpfd, "w+");
				if (outfile)
				{

					bool body = false;
					int lines = 0;
					int maxlines = 0;
					int bytes = 0;

					while( !m_bCancelDownload )
					{
					
						// Throttle the bandwidth
						while( bThrottleDownloadRate && g_pNZBQueueCoordinator->CalcCurrentDownloadSpeed() > g_fDownloadRate )
						{
							usleep( 500 * 1000 );
						}

						// Global Pausing??
						while( g_bPause )
						{
							usleep( 1000 * 1000 );
						}

						line = pConnection->ReadLine();

						//detect end of article
						if ((!line) || (!strcmp(line, ".\r\n")) || (!strcmp(line, ".\n")))
						break;

						// Did we meet an unexpected need for authorization?
						if( !strncmp(line, "480", 3) )
						{
							pConnection->authenticate();
						}

						//detect lines starting with "." (marked as "..")
						if (!strncmp(line, "..", 2))
							line++;
                      
						// detect Lines-header
						if ((!maxlines) && (!strncmp(line, "Lines: ",7)))
							maxlines = atoi(line+7);
                      
						// count lines in body
						if (body)
							lines++;
                      
						// detect end og header
						if ((!body) && ((!strcmp(line,"\r\n")) || (!strcmp(line,"\n"))))
						{
							body = true;
						}
                      
						int wrcnt = fwrite(line, 1, strlen(line), outfile);
						if (wrcnt > 0)
						{
							bytes += wrcnt;
						}
                                            
						if ((lines % 100 == 0) || (lines == maxlines))
						g_pNZBQueueCoordinator->JobProgress( m_pJob, pConnection, bytes, lines, maxlines );
					}
                  
					fflush(outfile);
					fclose(outfile);
                  
					if (line)
					{
						g_pNZBQueueCoordinator->JobFinished( m_pJob, pConnection );
						MultiServerPool::getInstance()->freeConnection(level ,pConnection);
						char* resFilename = Decoder::getInstance()->decode(tmpfilename, m_szTempDir);
                      
                      if (resFilename)
                        {
                          info("Decoded %s", resFilename);
                          
                          m_pJob->setResultFilename(resFilename);
                          m_pJob->setStatus(JOBSTATUS_FINISHED);
                          
                          free(resFilename);
                          remove(tmpfilename);
                          break;
                        }
                      else 
                        {
                          error("Decoding %s failed", m_pJob->getMessageID());
                          // possibly permanent error, abort
                          m_pJob->setStatus(JOBSTATUS_FAILED);
                          break;
                        }
                    }
                  else
                    {
                      warn("Unexpected end of %s", m_pJob->getMessageID());
                    }

                  remove(tmpfilename);

                }
              else
                {
                  warn("Could not create file %s", tmpfilename);
                }
          
            }
          else
            {
              if (!answer)
                warn("%s @ %s failed: Connection closed by remote host.", m_pJob->getMessageID(), pConnection->GetServer()->GetHost());
              else
                warn("%s @ %s failed (%s)", m_pJob->getMessageID(), pConnection->GetServer()->GetHost(), answer);
              g_pNZBQueueCoordinator->JobFinished( m_pJob, pConnection );
              
            }
        }
      else
        {
          //group-change failed. Ignore this, too noisy to report
        }
      

      MultiServerPool::getInstance()->freeConnection(level,pConnection);
      g_pNZBQueueCoordinator->JobFinished( m_pJob, pConnection );
      
      level++;
    }

  this->stop();
}
