/*
 *  This file if part of nzbget
 *
 *  Copyright (C) 2004  Sven Henkel <sidddy@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Revision: 1.13 $
 * $Date: 2004/08/10 19:04:35 $
 *
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <vector>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <fcntl.h>

#include "FileDownloader.h"
#include "ArticleDownloader.h"
#include "ArticleJob.h"
#include "global.h"
#include "Options.h"
#include "Observable.h"
#include "MultiServerPool.h"

extern Options* options;

FileDownloader::FileDownloader(DownloadJob* pJob, char* tempdir, char* destdir)
{
	m_bCancelDownload = false;
	
	m_pJob = pJob;
	m_ArticleJobs.clear();
	m_ArticleDownloads.clear();
	
	if( tempdir )
	{
		m_szTempDir = (char*) malloc(strlen(tempdir) + 2);
		memset(m_szTempDir, 0, strlen(tempdir) + 2);
		memcpy(m_szTempDir, tempdir, strlen(tempdir));
		if (m_szTempDir[strlen(m_szTempDir)-1] != '/')
		{
			m_szTempDir[strlen(m_szTempDir)] = '/';
		}
	}
  	else
	{
		m_szTempDir = strdup("./");
	}

	if (destdir)
	{
		m_szDestDir = (char*) malloc(strlen(destdir) + 2);
		memset(m_szDestDir, 0, strlen(destdir) + 2);
		memcpy(m_szDestDir, destdir, strlen(destdir));
		
		if (m_szDestDir[strlen(m_szDestDir)-1] != '/')
		{
			m_szDestDir[strlen(m_szDestDir)] = '/';
		}
	}
	else
	{
		m_szDestDir = strdup("./");
	}

	// Ensure the destination dir is created
	mkdir(m_szDestDir, S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);

	// Ensure the tmp dir is created
	mkdir(m_szTempDir, S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
}

FileDownloader::~FileDownloader()
{
	if( m_szTempDir )
		free(m_szTempDir);
	
	if( m_szDestDir )
		free(m_szDestDir);
	
	for( unsigned int i = 0; i < m_ArticleJobs.size(); i++ )
	{
		if( m_ArticleJobs[i] )
		{
			delete m_ArticleJobs[i];
			m_ArticleJobs[i] = NULL;
		}
	}
	
	for( unsigned int i = 0; i < m_ArticleDownloads.size(); i++ )
	{
		if( m_ArticleDownloads[i] )
		{
			delete m_ArticleDownloads[i];
			m_ArticleJobs[i] = NULL;
		}
	}
}


void FileDownloader::start()
{

	if( IsDupe() )
	{
		m_pJob->setStatus(JOBSTATUS_FINISHED);
		warn("File with subject \"%s\" seems to be duplicate, skipping.",m_pJob->getSubject());
		return;
	}

	for (unsigned int i = 0; i < m_pJob->getMessageIDs()->size(); i++)
	{
		if ((*(m_pJob->getMessageIDs()))[i] != NULL)
		{
			//start only when level-0-connection is available
			NNTPConnection* conn = MultiServerPool::getInstance()->getConnection(0);
			MultiServerPool::getInstance()->freeConnection(0, conn);
	
			ArticleJob* cjob = new ArticleJob((*(m_pJob->getMessageIDs()))[i]);
	
			for (unsigned int j = 0; j < m_pJob->getGroups()->size(); j++)
			{
				cjob->getGroups()->push_back((*(m_pJob->getGroups()))[j]);
			}
	
			cjob->addObserver(this);
			m_ArticleJobs.push_back(cjob);
		
			ArticleDownloader* dl = new ArticleDownloader(cjob, m_szTempDir);
	
			m_ArticleDownloads.push_back(dl);
		}
	}
	
	for (unsigned int i = 0; i < m_ArticleDownloads.size(); i++)
	{
		//start only when free connections are available for every level
		int level = 0;

		while( true )
		{
			NNTPConnection* conn = MultiServerPool::getInstance()->getConnection(level);
			if (conn)
			{
				MultiServerPool::getInstance()->freeConnection(level, conn);
			}
			else
			{
				break;
			}
			level++;
		}

		if( m_bCancelDownload )
		{
			return;
		}
	
		m_ArticleDownloads[i]->start();
		usleep(100*1000);
	}
}

void FileDownloader::notify(Observable* caller)
{
	debug("FileDownloader::notify(Observable* caller)");
	
	// all child-jobs ready?
	for (unsigned int i = 0; i < m_ArticleJobs.size(); i++)
	{
		if((m_ArticleJobs[i]->getStatus() != JOBSTATUS_FINISHED)&& (m_ArticleJobs[i]->getStatus() != JOBSTATUS_FAILED))
		{
			return;
		}
	}
	
	// all ready!
	
	// determine filename
	
	char* basename = NULL;
	int maxpart = 0;
	
	for (unsigned int i = 0; i < m_ArticleJobs.size(); i++)
	{
		char* partfn = m_ArticleJobs[i]->getResultFilename();
		if (partfn)
		{
			char* lastdot = strrchr(partfn, '.');
	
			if (lastdot)
			{
				char* localbasename = (char*) malloc(lastdot-partfn+1);
				memset(localbasename, 0, lastdot-partfn+1);
				memcpy(localbasename, partfn, lastdot-partfn);
	
				if (strncmp(localbasename, "nzbgettemp-unknown-", 19))
				{
					if (basename)
					{
						if (strcmp(basename, localbasename))
						{
							error("Filenames differ in parts of post!");
							m_pJob->setStatus(JOBSTATUS_FAILED);
							return;
						}
						else
						{
							free(localbasename);
						}
					}
					else
					{
						basename = localbasename;
					}
				}
	
				int localpart = atoi(lastdot+1);
				if (localpart > maxpart)
				{
					maxpart = localpart;
				}
			}
		}
	}
	
	if (!basename)
	{
		//ooh, something f*cked up, so let's create our own name
		basename = (char*) malloc(1024);
		snprintf(basename, 1024, "nzbget-unknown-XXXXXX");
		mktemp(basename);
	}
	
	// search childs for unnamed parts
	for (unsigned int i = 0; i < m_ArticleJobs.size(); i++)
	{
		char* partfn = m_ArticleJobs[i]->getResultFilename();
		if (partfn)
		{
			if (!strncmp(partfn, "nzbgettemp-unknown-", 19))
			{
				char oldfn[1024];
				char newfn[1024];
	
				snprintf(oldfn, 1024, "%s%s", m_szTempDir, partfn);
				snprintf(newfn, 1024, "%s%s.%d", m_szTempDir, basename, i+1);
	
				rename(oldfn, newfn);
	
				if ((signed int)i+1 > maxpart)
					maxpart = i+1;
			}
		}
	}
	
	FILE* outfile;
	
	char ofn[1024];
	snprintf(ofn, 1024, "%s%s",m_szDestDir, basename);
	
	
	// prevent overwriting existing files
	struct stat statbuf;
	int dupcount = 0;
	while (!stat(ofn, &statbuf))
	{
		dupcount++;
		snprintf(ofn, 1024, "%s%s_duplicate%d",m_szDestDir, basename, dupcount);
	}
	
	outfile = fopen(ofn, "w+");
	
	if (!outfile)
	{
		error("Could not create file %s!", basename);
		m_pJob->setStatus(JOBSTATUS_FAILED);
		return;
	}
	
	
	bool complete = true;
	
	for (int i = 1; i <= maxpart; i++)
	{
		FILE* infile;
	
		char fn[1024];
		snprintf(fn, 1024, "%s%s.%d", m_szTempDir, basename, i);
	
		infile = fopen(fn, "r");
	
		if (infile)
		{
			char buffer[10240];
	
			int cnt = 10240;
	
			while (cnt == 10240)
			{
				cnt = fread(buffer, 1, 10240, infile);
				fwrite(buffer, 1, cnt, outfile);
			}
	
			fclose(infile);
			remove(fn);
		}
		else
		{
			complete = false;
		}
	}
	
	fclose(outfile);
	
	if (complete)
	{
		m_pJob->setStatus(JOBSTATUS_FINISHED);
	//      info("Decoded %s", ofn);
	}
	else
	{
		char brokenfn[1024];
		snprintf(brokenfn, 1024, "%s_broken", ofn);
		rename(ofn, brokenfn);
		m_pJob->setStatus(JOBSTATUS_FAILED);
		warn("%s is incomplete!", ofn);
	}
}


bool FileDownloader::IsRunning()
{
	for (unsigned int i = 0; i < m_ArticleDownloads.size(); i++)
	{
		if (m_ArticleDownloads[i]->isRunning())
		{
			return true;
		}
	}
	return false;
}


bool FileDownloader::IsDupe()
{

	if (strcmp(options->getOption(OPTION_DUPECHECK, "yes"), "yes"))
	{
		return false;
	}
	
	if ((!m_pJob) || (!m_pJob->getSubject()) || (!m_szDestDir))
	{
		return false;
	}
	
	
	DIR *dir=opendir(m_szDestDir);
	struct dirent *entry;
	if (!dir)
	{
		return false;
	}
	
	std::vector< char* > existing;
	
	existing.clear();
		
	while ((entry=readdir(dir)))
	{
		if (!strcmp(entry->d_name, ".."))
			continue;
	
		if (!strcmp(entry->d_name,"."))
			continue;
	
		if (entry->d_name[0] == '.')
			continue;
	
		// do not consider short filenames, too many false positives!!
		if ((strlen(entry->d_name) < 6) && (strstr(entry->d_name, ".") == NULL))
			continue;
	
		existing.push_back(strdup(entry->d_name));
	}
	
	closedir(dir);
	
	// mmkay, got the names of existing files, now make substring-tests
	// with subject
	
	bool res = false;
	
	for (unsigned int i = 0; i < existing.size(); i++)
	{
		if (strstr(m_pJob->getSubject(), existing[i]))
		{
			// found a match!
			res = true;
		}
		free(existing[i]);
		existing[i] = NULL;
	}
	
	existing.clear();
	return res;
}

void FileDownloader::CancelDownload()
{
	m_bCancelDownload = true;

	std::vector<ArticleDownloader*>::iterator it;
	for( it=m_ArticleDownloads.begin(); it!=m_ArticleDownloads.end(); it++ )
	{
		(*it)->CancelDownload();
	}
}

