/*
 *  This file if part of nzbget
 *
 *  Copyright (C) 2004  Sven Henkel <sidddy@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Revision: 1.7 $
 * $Date: 2004/08/10 19:04:35 $
 *
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <errno.h>
#include <signal.h>

#include <libxml/parser.h>

#include "MultiServerPool.h"
#include "global.h"
#include "NZBQueue.h"
#include "Options.h"
#include "Coordinator.h"
#include "Frontend.h"
#include "ColoredFrontend.h"
#include "NCursesFrontend.h"
#include "NZBQueueCoordinator.h"
#include "NZBQueueMessageControl.h"

// Externs
extern Frontend* frontend;
extern Options* options;
extern int errno;
extern bool g_bUseLog;
extern int g_iProgramPID;

// Prototypes
void SigProc( int );
void SendNZBFileName( const char* szName );
void Cleanup();
bool CreateNZBCharBufferFromFilename( const char* szFileName, char** pBuffer, int* pBufferLength );

NZBQueueCoordinator* g_pNZBQueueCoordinator = NULL;

/*
 * Main loop
 */
int main(int argc, char *argv[])
{
	// Init options & get the name of the .nzb file
	debug( "options parsing" );
	options						= new Options(argc, argv);
	bool bServerMode			= !strcmp( options->getOption( OPTION_SERVER, NULL ), "yes" );
	bool bStandaloneMode		= !strcmp( options->getOption( OPTION_STANDALONE, NULL ), "yes" );
	const char* szDocumentName	= options->getOption( OPTION_NZBFILE, NULL );
	g_iProgramPID				= getpid();
//	const char* szBinPath		= options->getOption( OPTION_BINPATH, NULL );
	
	// Install a signal handler
	signal( SIGINT, SigProc );
	
	//------------------------------------------------
	// Not in standalone-mode
	//------------------------------------------------
	if( !bStandaloneMode && !bServerMode )
	{
		SendNZBFileName( szDocumentName );
		
		return 0;
	}

	// Create the queue coordinator
	if( !g_pNZBQueueCoordinator )
	{
		g_pNZBQueueCoordinator = new NZBQueueCoordinator( options );
	}

	//------------------------------------------------
	// Not using server-mode
	//------------------------------------------------
	if( !bServerMode || bStandaloneMode )
	{
		NZBQueue* pQueue = g_pNZBQueueCoordinator->AddNZBFileToQueue( szDocumentName, false );

		if( !pQueue )
		{
			abort("FATAL ERROR: Parsing NZB-document %s failed!!\n\n", szDocumentName ? szDocumentName : "N/A" );
	    
			return -1;
		}
	}

	// Create the front-end
	if( !strcmp(options->getOption(OPTION_OUTPUT, NULL), "curses") )
	{
		frontend = new NCursesFrontend();
	}
	else if(!strcmp(options->getOption(OPTION_OUTPUT, NULL), "colored") )
	{
		frontend = new ColoredFrontend();
	}
	else
	{
		frontend = new Frontend();
	}

	g_pNZBQueueCoordinator->Start();
		
	printf("Download finished.                         \n\n");
	
	Cleanup();
	
	return 0;
}

/*
 * Signal handler for SIGINT - catch ctrl+c and cleanup IPC and other stuff
 */
void SigProc( int iSignal )
{
	Cleanup();

	signal( SIGINT, SIG_DFL ); // Reset the signal handler
	
	kill( getpid(), SIGINT );  // send a kill signal
}

void Cleanup()
{
	if( frontend )
	{
		delete frontend;
		frontend = NULL;
	}
	
	if( g_pNZBQueueCoordinator )
	{
		delete g_pNZBQueueCoordinator;
		g_pNZBQueueCoordinator = NULL;
	}
}

/*
 * Sends a message to the running nzbget process.
 */
void SendNZBFileName( const char* szName )
{
	// Read the file into the buffer
	char* szBuffer	= NULL;
	int iLength		= 0;
	CreateNZBCharBufferFromFilename( szName, &szBuffer, &iLength );

	// Create a connection to the server
	const char* szNetAddress	= options->getOption( OPTION_SERVERIP, NULL );
	int iPort					= atoi( options->getOption( OPTION_SERVERPORT, NULL ) );
	const char* szPassword		= options->getOption( OPTION_SERVERPASSWORD, NULL );
	NetAddress* pNetAddress		= new NetAddress( szNetAddress, iPort );
	Connection* pConnection		= new Connection( pNetAddress );

	if( pConnection->Connect() >= 0 )
	{
		// Send a "new download" request
		SNZBQueueRequest Request;
		Request.m_iId	= 'nzbg';
		Request.m_iType = NZBMessageRequest::eNewDownload;
		if( strlen( szName ) < NZBREQUESTFILENAMESIZE )
		{
			strcpy( Request.m_szFileName, szName );
		}
		else
		{
			memcpy( Request.m_szFileName, szName, NZBREQUESTFILENAMESIZE-1 );
			Request.m_szFileName[ NZBREQUESTFILENAMESIZE-1 ] = 0;
		}
		strncpy( Request.m_szPassword, szPassword, NZBREQUESTPASSWORDSIZE-1 );

		if( pConnection->Send( (char*) (&Request), sizeof( Request )  ) < 0 )
		{
			perror( "pConnection->Write" );
			return;
		}
		else
		{
			pConnection->Write( szBuffer );
			printf( "Request send to %s (port %i)\n", szNetAddress, iPort );

/*			Doesnt seem to work
			char* szAnswer = pConnection->ReadLine();
			if( szAnswer )
			{
				printf( "nzbserver returned: %s\n", szAnswer );
			}
*/
			pConnection->Disconnect();
		}
	}
	else
	{
		printf( "Unable to send request to nzbserver at %s (port %i)\n", szNetAddress, iPort );
	}

	// Cleanup
	if( szBuffer )
	{
		free( szBuffer );
	}
}

bool CreateNZBCharBufferFromFilename( const char* szFileName, char** pBuffer, int* pBufferLength )
{
	FILE* pFile = fopen( szFileName, "r" );

	if( pFile==NULL )
	{
		exit( -1 );
		return false;
	}

	// obtain file size.
	fseek (pFile , 0 , SEEK_END);
	int iSize  = ftell (pFile);
	rewind (pFile);

	// allocate memory to contain the whole file.
	*pBuffer = (char*) malloc ( iSize+1 );
	if( *pBuffer == NULL)
	{
		exit( -1 );
		return false;
	}
	
	// copy the file into the buffer.
	fread( *pBuffer, 1, iSize, pFile );

	fclose (pFile);

	(*pBuffer)[ iSize ] = 0;

	*pBufferLength = iSize;

	return true;
}
